/**
 * @file    shellmatta_transport.c
 * @brief   transport layer functions of shellmatta
 * @author  Simon Fischer <fischer.simon.1991@gmail.com>
 */

#include "shellmatta_transport.h"
#include "shellmatta.h"
#include "shellmatta_crc.h"
#include <string.h>


/**
 * @brief           Initializes the transportLayerInst
 * @param[in, out]  transportLayer  transport layer instance to work on
 * @param[in]       writeFct        function pointer to output function
 * @return          errorcode       #SHELLMATTA_OK
 */
shellmatta_retCode_t shellmatta_transport_init( shellmatta_transport_layer_t    *transportLayer,
                                                shellmatta_write_t              writeFct)
{
    memset(transportLayer, 0u, sizeof(shellmatta_transport_layer_t));
    transportLayer->originalWrite = writeFct;

    return SHELLMATTA_OK;
}

/**
 * @brief           Resets all values of tranportLayerInst except for sequence counters
 * @param[in, out]  transportLayer  transport layer instance to work on
 * @return          errorcode       #SHELLMATTA_OK
 */
shellmatta_retCode_t shellmatta_transport_reset(shellmatta_transport_layer_t *transportLayer)
{
    uint8_t h2s_sequenceCnt = transportLayer->h2s_sequenceCnt;
    uint8_t s2h_sequenceCnt = transportLayer->s2h_sequenceCnt;
    memset(transportLayer, 0u, sizeof(shellmatta_transport_layer_t));

    transportLayer->h2s_sequenceCnt = h2s_sequenceCnt;
    transportLayer->s2h_sequenceCnt = s2h_sequenceCnt;

    return SHELLMATTA_OK;
}

/**
 * @brief           processes the passed amount of data
 * @param[in, out]  transportLayer  transport layer instance to work on
 * @param[in]       data            pointer to input data to process
 * @return          errorcode       #SHELLMATTA_OK\n
 *                                  #SHELLMATTA_ERROR   in case of crc error
 */
shellmatta_retCode_t shellmatta_transport_process(shellmatta_transport_layer_t   *transportLayer,
                                                  char                           *data,
                                                  uint32_t                       *length)
{
    shellmatta_retCode_t ret = SHELLMATTA_OK;


    //TODO
    (void)length;

    switch (transportLayer->state)
    {
        /* look for start of header */
    case STATE_GET_SOH:
        /* if start of header is found, continue transport layer fsm */
        if (SHELLMATTA_START_OF_HEADER == *data)
        {
            transportLayer->state = STATE_GET_PROTOCOL_VERSION;
            transportLayer->active = true;
        }
        else if (0x00 != *data)
        {
            /* if SOH is not found, handle data as manual input */
            transportLayer->state = STATE_MANUAL_INPUT;
            transportLayer->active = false;
        }
        break;

    case STATE_MANUAL_INPUT:
        break;

    /* look for protocol version */
    case STATE_GET_PROTOCOL_VERSION:
        /* check protocol version */
        if (PROTOCOL_VERSION == *data)
        {
            transportLayer->state = STATE_GET_PACKET_TYPE;
        }
        else
        {
            /* reset fsm in case of wrong protocol version */
            transportLayer->state = STATE_GET_SOH;
            transportLayer->active = false;
        }
        break;

    /* read packet type */
    case STATE_GET_PACKET_TYPE:
        transportLayer->state = STATE_GET_PAYLOAD_LENGTH;
        break;

    /* look for length of payload */
    case STATE_GET_PAYLOAD_LENGTH:
        transportLayer->state = STATE_GET_SOURCE;
        break;

    /* look for source */
    case STATE_GET_SOURCE:
        transportLayer->state = STATE_GET_DESTINATION;
        break;

    /* look for destination */
    case STATE_GET_DESTINATION:
        transportLayer->state = STATE_GET_H2S_SEQUENCE_CNT;
        break;

    /* look for h2s sequence counter*/
    case STATE_GET_H2S_SEQUENCE_CNT:
        transportLayer->state = STATE_GET_S2H_SEQUENCE_CNT;
        break;

    /* look for s2h sequence counter */
    case STATE_GET_S2H_SEQUENCE_CNT:
        if (transportLayer->packetType == PACKET_SEQ_CNT_REQUEST)
        {
            transportLayer->state = STATE_GET_CRC;
        }
        else
        {
            transportLayer->state = STATE_GET_PAYLOAD;
        }
        break;

    /* read payload for previously read length of payload */
    case STATE_GET_PAYLOAD:
        if (    (transportLayer->payloadLength <= transportLayer->payloadCounter)
            &&  (true == transportLayer->active))
        {
            transportLayer->state = STATE_GET_CRC;
        }
        break;

    /* read crc32 for four bytes */
    case STATE_GET_CRC:
        /* if crc32 is read, start computing */
        if (SHELLMATTA_LENGTH_CRC <= transportLayer->crcCounter)
        {
            /* for crc computation only */
            char crcdata[SHELLMATTA_PAYLOAD_MAXLENGTH + SHELLMATTA_HEADER_LENGTH];
            memset(crcdata, 0u, SHELLMATTA_PAYLOAD_MAXLENGTH + SHELLMATTA_HEADER_LENGTH);
            crcdata[SHELLMATTA_POS_SOH]         = SHELLMATTA_START_OF_HEADER;
            crcdata[SHELLMATTA_POS_PROT_VER]    = transportLayer->protocolVersion;
            crcdata[SHELLMATTA_POS_PACKET_TYPE] = transportLayer->packetType;
            crcdata[SHELLMATTA_POS_PAYLOAD_LEN] = transportLayer->payloadLength;
            crcdata[SHELLMATTA_POS_SRC]         = transportLayer->source;
            crcdata[SHELLMATTA_POS_DST]         = transportLayer->destination;
            crcdata[SHELLMATTA_POS_CNT_H2S]     = transportLayer->packetSequenceCounter_h2s;
            crcdata[SHELLMATTA_POS_CNT_S2H]     = transportLayer->packetSequenceCounter_s2h;
            strncat(&crcdata[SHELLMATTA_POS_PAYLOAD], transportLayer->payloadBuffer, transportLayer->payloadLength);

            uint32_t refCrc;

            refCrc = crc32Calc(crcdata, SHELLMATTA_HEADER_LENGTH + transportLayer->payloadLength);

            /* if crc is correct, further handling of data depends on type of packet */
            if (transportLayer->crc32 == refCrc)
            {
                /* reset fsm by default */
                transportLayer->state = STATE_GET_SOH;

                /* crc is correct */
                transportLayer->h2s_sequenceCnt++;

                switch (transportLayer->packetType)
                {
                case PACKET_DATA:
                    transportLayer->state = STATE_PROCESS_PAYLOAD;
                    break;
                
                case PACKET_SEQ_CNT_REQUEST:
                    /* change packet type to response */
                    transportLayer->packetType = PACKET_SEQ_CNT_RESPOND;
                    /* send out packet with no payload */
                    shellmatta_transport_write(transportLayer, "", 0u);
                    break;

                case PACKET_SEQ_CNT_RESPOND:
                    break;
                
                case PACKET_MAX_BUFFERSIZE_REQUEST:
                    break;

                case PACKET_MAX_BUFFERSIZE_RESPOND:
                    break;
                
                case PACKET_SEARCH_DEVICE_REQUEST:
                    break;
                
                case PACKET_SEARCH_DEVICE_RESPOND:
                    break;

                case PACKET_SET_ADDRESS_REQUEST:
                    break;

                case PACKET_SET_ADDRESS_RESPOND:
                    break;

                /* wrong packet type */
                default:
                    /* undo sequence counter increment */
                    transportLayer->h2s_sequenceCnt--;
                    break;
                }
            }
            else
            {
                transportLayer->state = STATE_GET_SOH;
                /* crc is incorrect */
                ret = SHELLMATTA_ERROR;
            }

            /* reset crc variables */
            transportLayer->crcCounter = 0u;
            transportLayer->crc32 = 0u;
        }
        break;

    /* start to process payload as if it was entered manually */
    case STATE_PROCESS_PAYLOAD:
        transportLayer->state = STATE_GET_SOH;
        transportLayer->protocolVersion = 0u;
        transportLayer->packetType      = PACKET_DATA;
        transportLayer->payloadLength   = 0u;
        transportLayer->source          = 0u;
        transportLayer->destination     = 0u;
        transportLayer->crc32           = 0u;
        transportLayer->payloadCounter  = 0u;
        transportLayer->crcCounter      = 0u;
        transportLayer->active = false;
        memset(transportLayer->payloadBuffer, 0u, SHELLMATTA_PAYLOAD_MAXLENGTH);
        break;
    
    default:
        break;
    }

    return ret;
}

/**
 * @brief       Wrapper function for the write-function of shellmatta handle
 * 
 * This function is used to transmit data with the tranport layer protocol.\n
 * The input data is copied into a new array along with the header and crc32.\n
 * The resulting buffer is the forwarded to the original write function.\n
 * 
 * @note        If length of data exceeds #UINT8_MAX, data is sent in multiple packets.\n
 * 
 * @param[in, out]  transportLayer  transport layer instance to work on
 * @param[in]       data            pointer to input data to process
 * @return          errorcode       #SHELLMATTA_OK
 */
shellmatta_retCode_t shellmatta_transport_write(shellmatta_transport_layer_t *transportLayer,
                                                const char* data,
                                                uint32_t length)
{
    shellmatta_retCode_t ret = SHELLMATTA_OK;
    
    /* create buffer for data + header + crc with maximum size */
    uint8_t outputBuffer[   SHELLMATTA_HEADER_LENGTH +      /* length of header */
                            SHELLMATTA_PAYLOAD_MAXLENGTH +  /* max length of payload */
                            SHELLMATTA_LENGTH_CRC];         /* length of crc */
    
    uint16_t outputBufferSize = sizeof(outputBuffer);
    uint32_t outPayloadLength = length;

    /* handle data with length bigger than max length */
    uint16_t processedLength = 0u;
    uint16_t splitLength = 0u;
    uint16_t splitPayloadLength = 0u;
    uint16_t restOfPayload = 0u;

    /* foot-controlled loop to send packets without payload length */
    do
    {
        /* compute length of next payload split */
        restOfPayload = (outPayloadLength - processedLength);
        if (restOfPayload > SHELLMATTA_PAYLOAD_MAXLENGTH)
        {
            splitLength = outputBufferSize;
        }
        else
        {
            splitLength = (restOfPayload + SHELLMATTA_HEADER_LENGTH + SHELLMATTA_LENGTH_CRC) % outputBufferSize;
        }
        splitPayloadLength = splitLength - SHELLMATTA_HEADER_LENGTH - SHELLMATTA_LENGTH_CRC;

        /* fill buffer with header data */
        outputBuffer[SHELLMATTA_POS_SOH] = SHELLMATTA_START_OF_HEADER;                  /* start of header                      */
        outputBuffer[SHELLMATTA_POS_PROT_VER] = PROTOCOL_VERSION;                       /* protocol version                     */
        outputBuffer[SHELLMATTA_POS_PACKET_TYPE] = transportLayer->packetType;       /* packet type                          */
        outputBuffer[SHELLMATTA_POS_PAYLOAD_LEN] = splitPayloadLength;                  /* payload length                       */
        outputBuffer[SHELLMATTA_POS_SRC] = 0x00u;                                       /* source                               */
        outputBuffer[SHELLMATTA_POS_DST] = 0x00u;                                       /* destination                          */
        outputBuffer[SHELLMATTA_POS_CNT_H2S] = transportLayer->h2s_sequenceCnt;      /* sequence counter host to shellmatta  */
        outputBuffer[SHELLMATTA_POS_CNT_S2H] = ++transportLayer->s2h_sequenceCnt;    /* sequence counter shellmatta to host  */

        /* skip copying of payload in case of sequence counter response */
        if (transportLayer->packetType != PACKET_SEQ_CNT_RESPOND)
        {
            memcpy(&outputBuffer[SHELLMATTA_POS_PAYLOAD], data + processedLength, splitPayloadLength);
        }

        uint32_t outCrc = crc32Calc((char*) outputBuffer, SHELLMATTA_HEADER_LENGTH + splitPayloadLength);

        /* append crc to end of payload */
        outputBuffer[splitPayloadLength + SHELLMATTA_HEADER_LENGTH] = (uint8_t)(outCrc >> SHELLMATTA_SHIFT_BY_24);
        outputBuffer[splitPayloadLength + SHELLMATTA_HEADER_LENGTH + SHELLMATTA_SHIFT_BY_1] = (uint8_t)(outCrc >> SHELLMATTA_SHIFT_BY_16);
        outputBuffer[splitPayloadLength + SHELLMATTA_HEADER_LENGTH + SHELLMATTA_SHIFT_BY_2] = (uint8_t)(outCrc >> SHELLMATTA_SHIFT_BY_8);
        outputBuffer[splitPayloadLength + SHELLMATTA_HEADER_LENGTH + SHELLMATTA_SHIFT_BY_3] = (uint8_t)(outCrc);

        /* use original write function to send full buffer */
        ret = transportLayer->originalWrite((char*) outputBuffer, splitLength);
        processedLength += splitPayloadLength;
    }
    while (processedLength < outPayloadLength);
    return ret;
}
